from django.conf.urls import url
from ucamwebauth.views import raven_login, raven_logout, raven_return

urlpatterns = [
    url(r'^accounts/login/$', raven_login, name='raven_login'),
    url(r'^accounts/logout/$', raven_logout, name='raven_logout'),
    url(r'^raven_return/$', raven_return, name='raven_return'),
]



from django.contrib import messages
from django.http import HttpResponseServerError, HttpResponseForbidden
from django.template.loader import get_template
from ucamwebauth import MalformedResponseError, InvalidResponseError, PublicKeyNotFoundError, UserNotAuthorised, \
    OtherStatusCode


class DefaultErrorBehaviour():
    """ A middleware that catches django-ucamwebauth exceptions and show HTTP 500 or HTTP 403 error messages,
    depending of the error. Furthermore, it uses templates that can be rewritten by a developer.
    """
    def process_exception(self, request, exception):
        if exception.__class__ == MalformedResponseError or \
                exception.__class__ == InvalidResponseError or \
                exception.__class__ == OtherStatusCode or \
                exception.__class__ == PublicKeyNotFoundError:
            template = get_template("ucamwebauth_500.html")
            messages.error(request, str(exception))
            return HttpResponseServerError(template.render({}, request))
        elif exception.__class__ == UserNotAuthorised:
            template = get_template("ucamwebauth_403.html")
            messages.error(request, str(exception))
            return HttpResponseForbidden(template.render({}, request))



from base64 import b64encode
from datetime import datetime, timedelta
from django.conf import settings
try:
    from urlparse import urlparse, parse_qs
    from urllib import unquote, urlencode
except ImportError:
    from urllib.parse import urlparse, parse_qs, unquote, urlencode
import sys
from OpenSSL.crypto import load_privatekey, FILETYPE_PEM, sign
import requests
from django.test import TestCase, RequestFactory
from django.test.client import Client
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User
from ucamwebauth import InvalidResponseError, MalformedResponseError, UserNotAuthorised, RavenResponse, \
    PublicKeyNotFoundError
from ucamwebauth.exceptions import OtherStatusCode
from ucamwebauth.utils import get_next_from_wls_response, get_return_url
from ucamwebauth.backends import RavenAuthBackend

RAVEN_TEST_USER = 'test0001'
RAVEN_TEST_PWD = 'test'
RAVEN_NEW_USER = 'test0002'
RAVEN_FORLIVE_USER = 'test0500'

GOOD_PRIV_KEY_PEM = """-----BEGIN RSA PRIVATE KEY-----
MIICWwIBAAKBgQC4RYvbSGb42EEEXzsz93Mubo0fdWZ7UJ0HoZXQch5XIR0Zl8AN
aLf3tVpRz4CI2JBUVpUjXEgzOa+wZBbuvczOuiB3BfNDSKKQaftxWKouboJRA5ac
xa3fr2JZc8O5Qc1J6Qq8E8cjuSQWlpxTGa0JEnbKV7/PVUFDuFeEI11e/wIDAQAB
AoGACr2jBUkXF3IjeAnE/aZyxEYVW7wQGSf9vzAf92Jvekyn0ZIS07VC4+FiPlqF
93QIFaJmVwVOAA5guztaStgtU9YX37wRPkFwrtKgjZcqV8ReQeC67bjo5v3Odht9
750F7mKWXctZrm0MD1PoDlkLvVZ2hDolHm5tpfP52jPvQ6ECQQDgtI4K3IuEVOIg
75xUG3Z86DMmwPmme7vsFgf2goWV+p4471Ang9oN7l+l+Jj2VISdz7GE7ZQwW6a1
IQev3+h7AkEA0e9oC+lCcYsMsI9vtXvB8s6Bpl0c1U19HUUWHdJIpluwvxF6SIL3
ug4EJPP+sDT5LvdV5cNy7nmO9uUd+Se2TQJAdxI2UrsbkzwHt7xA8rC60OWadWa8
4+OdaTUjcxUnBJqRTUpDBy1vVwKB3MknBSE0RQvR3canSBjI9iJSmHfmEQJAKJlF
49fOU6ryX0q97bjrPwuUoxmqs81yfrCXoFjEV/evbKPypAc/5SlEv+i3vlfgQKbw
Y6iyl0/GyBRzAXYemQJAVeChw15Lj2/uE7HIDtkqd8POzXjumOxKPfESSHKxRGnP
3EruVQ6+SY9CDA1xGfgDSkoFiGhxeo1lGRkWmz09Yw==
-----END RSA PRIVATE KEY-----
"""

BAD_PRIV_KEY_PEM = """-----BEGIN RSA PRIVATE KEY-----
MIICXQIBAAKBgQD5mkLpi7q6ROdu7khB3S9aanA0Zls7vvfGOmB80/yeylhGpsjA
jWen0VtSQke/NlEPGtO38tsV7CsuFnSmschvAnGrcJl76b0UOOHUgDTIoRxC6QDU
3claegwsrBA+sJEBbqx5RdXbIRGicPG/8qQ4Zm1SKOgotcbwiaor2yxZ2wIDAQAB
AoGBAPCgMpmLxzwDaUmcFbTJUvlLW1hoxNNYSu2jIZm1k/hRAcE60JYwvBkgz3UB
yMEh0AtLxYe0bFk6EHah11tMUPgscbCq73snJ++8koUw+csk22G65hOs51bVb7Aa
6JBe67oLzdtvgCUFAA2qfrKzWRZzAdhUirQUZgySZk+Xq1pBAkEA/kZG0A6roTSM
BVnx7LnPfsycKUsTumorpXiylZJjTi9XtmzxhrYN6wgZlDOOwOLgSQhszGpxVoMD
u3gByT1b2QJBAPtL3mSKdvwRu/+40zaZLwvSJRxaj0mcE4BJOS6Oqs/hS1xRlrNk
PpQ7WJ4yM6ZOLnXzm2mKyxm50Mv64109FtMCQQDOqS2KkjHaLowTGVxwC0DijMfr
I9Lf8sSQk32J5VWCySWf5gGTfEnpmUa41gKTMJIbqZZLucNuDcOtzUaeWZlZAkA8
ttXigLnCqR486JDPTi9ZscoZkZ+w7y6e/hH8t6d5Vjt48JVyfjPIaJY+km58LcN3
6AWSeGAdtRFHVzR7oHjVAkB4hutvxiOeiIVQNBhM6RSI9aBPMI21DoX2JRoxvNW2
cbvAhow217X9V0dVerEOKxnNYspXRrh36h7k4mQA+sDq
-----END RSA PRIVATE KEY-----
"""


def wls_response_escape(x):
    return x.replace('%', '%25').replace('!', '%21')


def create_wls_response(raven_ver='3', raven_status='200', raven_msg='',
                        raven_issue=datetime.utcnow().strftime('%Y%m%dT%H%M%SZ'),
                        raven_id='1347296083-8278-2',
                        raven_url=None,
                        raven_principal=RAVEN_TEST_USER, raven_ptags='current',
                        raven_auth='pwd', raven_sso='', raven_life='36000',
                        raven_params='', raven_kid='901',
                        raven_key_pem=GOOD_PRIV_KEY_PEM, raven_sig_input=True):
    """Creates a valid WLS Response as the Raven test server would
    using keys from https://raven.cam.ac.uk/project/keys/demo_server/
    """
    if raven_url is None:
        raven_url = (
            get_return_url(RequestFactory().get(reverse('raven_return'))))
    raven_pkey = load_privatekey(FILETYPE_PEM, raven_key_pem)

    # This is the data which is signed by Raven with their private key
    # Note data consists of full payload with exception of kid and sig
    # source: http://raven.cam.ac.uk/project/waa2wls-protocol-3.0.txt
    wls_response_data = [raven_ver, raven_status, raven_msg,
                         raven_issue, raven_id, raven_url,
                         raven_principal, raven_ptags, raven_auth,
                         raven_sso, raven_life, raven_params]

    data = '!'.join(wls_response_data)
    raven_sig = b64encode(sign(raven_pkey, data.encode(), 'sha1'))

    # Full WLS-Response also includes the Raven-variant b64encoded sig
    # and the requisite Key ID which has been used for the signing
    # process
    wls_response_data.append(raven_kid)
    if raven_sig_input:
        wls_response_data.append(raven_sig.decode().replace("+", "-").replace("/", ".").replace("=", "_"))
    else:
        wls_response_data.append('')

    return '!'.join(map(wls_response_escape, wls_response_data))


class RavenTestCase(TestCase):
    fixtures = ['users.json']

    def __init__(self, *args, **kwargs):
        self.client = Client()
        super(RavenTestCase, self).__init__(*args, **kwargs)

    def get_wls_response(self, raven_user=RAVEN_TEST_USER, raven_pwd=RAVEN_TEST_PWD, raven_ver='3',
                         raven_url=None, raven_desc='',
                         raven_aauth='pwd', raven_iact='', raven_msg='',
                         raven_params='', raven_fail='', cancel=False):
        # This request only test when raven_aauth is pwd and raven_iact is omitted
        if raven_url is None:
            raven_url = (
                get_return_url(RequestFactory().get(reverse('raven_return'))))
        if cancel:
            response = requests.post('https://demo.raven.cam.ac.uk/auth/authenticate2.html',
                                     {'userid': raven_user, 'pwd': raven_pwd, 'ver': raven_ver, 'url': raven_url,
                                      'params': raven_params, 'fail': raven_fail, 'cancel': 'Cancel'},
                                     allow_redirects=False)
        else:
            response = requests.post('https://demo.raven.cam.ac.uk/auth/authenticate2.html',
                                     {'userid': raven_user, 'pwd': raven_pwd, 'ver': raven_ver, 'url': raven_url,
                                      'params': raven_params, 'fail': raven_fail},
                                     allow_redirects=False)
        self.assertEqual(303, response.status_code)
        return unquote(response.headers['location']).split('WLS-Response=')[1]

    def test_login_raven_not_local(self):
        """Tests login of user via raven, not in database"""
        with self.settings(UCAMWEBAUTH_CREATE_USER=False):
            self.client.get(reverse('raven_return'), {'WLS-Response': self.get_wls_response(raven_user=RAVEN_NEW_USER)})
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_login_raven_local(self):
        """Tests login of user who exists in database"""
        with self.settings(UCAMWEBAUTH_CREATE_USER=False):
            self.client.get(reverse('raven_return'), {'WLS-Response': self.get_wls_response()})
        self.assertIn('_auth_user_id', self.client.session)

    def test_login_invalid_version_fails(self):
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': self.get_wls_response(raven_ver='4')})
        self.assertEqual(str(excep.exception), 'Unsupported version: 4')
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_login_invalid_version_fails_with_template(self):
        with self.settings(
                MIDDLEWARE_CLASSES=(
                    'django.contrib.sessions.middleware.SessionMiddleware',
                    'django.middleware.common.CommonMiddleware',
                    'django.middleware.csrf.CsrfViewMiddleware',
                    'django.contrib.auth.middleware.AuthenticationMiddleware',
                    'django.contrib.messages.middleware.MessageMiddleware',
                    'django.middleware.clickjacking.XFrameOptionsMiddleware',
                    'ucamwebauth.middleware.DefaultErrorBehaviour',
                )):
            response = self.client.get(reverse('raven_return'), {'WLS-Response': self.get_wls_response(raven_ver='4')})
        self.assertContains(response, 'Unsupported version: 4', status_code=500)
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_login_issue_future_fails(self):
        """Tests that Raven responses issued in the future fail validation"""
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response(
                                raven_issue=(datetime.utcnow() + timedelta(hours=1)).strftime('%Y%m%dT%H%M%SZ'))})
        self.assertEqual(str(excep.exception), 'The timestamp on the response is in the future')
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_wrong_status_code(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_status='100')})
        self.assertEqual(str(excep.exception), "Status returned not known")
        self.assertNotIn('_auth_user_id', self.client.session)
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_status='abc')})
        self.assertEqual(str(excep.exception), "Status code must be an integer, not abc")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_wrong_num_args(self):
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response().replace('!200!!', '!200!')})
        self.assertEqual(str(excep.exception), "Wrong number of parameters in response: expected 14, got 13")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_wrong_no_wlsresponse(self):
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'), {})
        self.assertEqual(str(excep.exception), "no WLS-Response")
        self.assertNotIn('_auth_user_id', self.client.session)
        with self.assertRaises(TypeError) as excep:
            RavenResponse()
        self.assertEqual(str(excep.exception), "no request supplied")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_login_issue_future_fails_with_template(self):
        """Tests that Raven responses issued in the future fail validation"""
        with self.settings(
                MIDDLEWARE_CLASSES=(
                    'django.contrib.sessions.middleware.SessionMiddleware',
                    'django.middleware.common.CommonMiddleware',
                    'django.middleware.csrf.CsrfViewMiddleware',
                    'django.contrib.auth.middleware.AuthenticationMiddleware',
                    'django.contrib.messages.middleware.MessageMiddleware',
                    'django.middleware.clickjacking.XFrameOptionsMiddleware',
                    'ucamwebauth.middleware.DefaultErrorBehaviour',
                )):
            response = self.client.get(reverse('raven_return'),
                                       {'WLS-Response': create_wls_response(
                                           raven_issue=(datetime.utcnow() +
                                                        timedelta(hours=1)).strftime('%Y%m%dT%H%M%SZ'))})
        self.assertContains(response, 'The timestamp on the response is in the future', status_code=500)
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_login_issue_too_old_fails(self):
        """Tests that Raven responses which are older than UCAMWEBAUTH_TIMEOUT are rejected"""
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response(
                                raven_issue=(datetime.utcnow() + timedelta(hours=-1)).strftime('%Y%m%dT%H%M%SZ'))})
        self.assertTrue(str(excep.exception).startswith('Response has timed out'))
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_wrong_date(self):
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response(raven_issue="error")})
        self.assertTrue(str(excep.exception).startswith("Issue time is not a valid time, got error"))
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_wrong_return_url(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response(raven_url="error")})
        self.assertTrue(str(excep.exception).startswith("The URL in the response does not match"))
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_username_when_not_status_200(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response(raven_status="510")})
        self.assertTrue(str(excep.exception).startswith("The username should not be present if the status code is "
                                                        "not 200"))
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_login_wrong_private_key_fails(self):
        """Tests that Raven responses with invalid key fail"""
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response(raven_key_pem=BAD_PRIV_KEY_PEM)})
        self.assertEqual(str(excep.exception), 'The signature for this response is not valid.')
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_create_raven_not_local_create_false(self):
        """When valid raven user authenticates, and UCAMWEBAUTH_CREATE_USER is
        false, user is not created in database"""
        with self.settings(UCAMWEBAUTH_CREATE_USER=False):
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': self.get_wls_response(raven_user=RAVEN_NEW_USER)})
            with self.assertRaises(User.DoesNotExist):
                User.objects.get(username=RAVEN_NEW_USER)
            self.assertNotIn('_auth_user_id', self.client.session)

    def test_raven_user_not_local_create_true(self):
        """When valid raven user authenticates, and UCAMWEBAUTH_CREATE_USER is true
        creates valid user in database"""
        with self.settings(UCAMWEBAUTH_CREATE_USER=True):
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': self.get_wls_response(raven_user=RAVEN_NEW_USER)})
            user = User.objects.get(username=RAVEN_NEW_USER)
            self.assertFalse(user.has_usable_password())

    def test_wrong_kid(self):
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_kid='error')})
        self.assertEqual(str(excep.exception), "kid parameter must be an integer, not error")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_empty_kid(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_kid='')})
        self.assertEqual(str(excep.exception), "kid must be present if signature is present")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_unknown_kid(self):
        with self.assertRaises(PublicKeyNotFoundError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_kid='100')})
        self.assertEqual(str(excep.exception), "The server do not have the public key corresponding to the key the web "
                                               "login service signed the response with")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_no_auth_no_sso(self):
        with self.assertRaises(MalformedResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_auth='',
                                                                                          raven_sso='')})
        self.assertEqual(str(excep.exception), "No authentication types supplied")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_unknown_sso(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_auth='',
                                                                                          raven_sso='card')})
        self.assertEqual(str(excep.exception), "The response used the wrong type of authentication (sso)")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_unknown_auth(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_auth='card',
                                                                                          raven_sso='')})
        self.assertEqual(str(excep.exception), "The response used the wrong type of authentication (auth)")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_empty_signature(self):
        with self.assertRaises(InvalidResponseError) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': create_wls_response(raven_sig_input=False)})
        self.assertEqual(str(excep.exception), "Signature must be present if status is 200")
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_logout_redirect_url(self):
        """Tests the logout redirection"""
        self.client.get(reverse('raven_return'), {'WLS-Response': self.get_wls_response()})
        self.assertIn('_auth_user_id', self.client.session)
        with self.settings(UCAMWEBAUTH_LOGOUT_REDIRECT='http://www.cam.ac.uk/'):
            response = self.client.get(reverse('raven_logout'), follow=True)
            self.assertEqual('http://www.cam.ac.uk/', response.redirect_chain[0][0])
            self.assertEqual(302, response.redirect_chain[0][1])

    def test_not_allow_raven_for_life(self):
        """Test Raven for life accounts credentials"""
        with self.assertRaises(UserNotAuthorised) as excep:
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': self.get_wls_response(raven_user=RAVEN_FORLIVE_USER)})
        self.assertEqual(str(excep.exception), 'Authentication successful but you are not authorised to access this '
                                               'site')

    def test_not_allow_raven_for_life_with_template(self):
        """Test Raven for life accounts credentials"""
        with self.settings(
                MIDDLEWARE_CLASSES=(
                    'django.contrib.sessions.middleware.SessionMiddleware',
                    'django.middleware.common.CommonMiddleware',
                    'django.middleware.csrf.CsrfViewMiddleware',
                    'django.contrib.auth.middleware.AuthenticationMiddleware',
                    'django.contrib.messages.middleware.MessageMiddleware',
                    'django.middleware.clickjacking.XFrameOptionsMiddleware',
                    'ucamwebauth.middleware.DefaultErrorBehaviour',
                )):
            response = self.client.get(reverse('raven_return'),
                                       {'WLS-Response': self.get_wls_response(raven_user=RAVEN_FORLIVE_USER)})
        self.assertContains(response, 'Authentication successful but you are not authorised to access this site',
                            status_code=403)

    def test_allow_raven_for_life(self):
        with self.settings(UCAMWEBAUTH_NOT_CURRENT=True, UCAMWEBAUTH_CREATE_USER=True):
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': self.get_wls_response(raven_user=RAVEN_FORLIVE_USER)})
            self.assertIn('_auth_user_id', self.client.session)

    def test_user_cancel_wls_auth(self):
        with self.assertRaises(OtherStatusCode) as excep:
            self.client.get(reverse('raven_return'), {'WLS-Response': self.get_wls_response(cancel=True)})
        self.assertEqual(str(excep.exception),
                         'The WLS returned status 410: The user cancelled the authentication request')
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_return_url(self):
        u = "http://elsewhere.example"
        with self.settings(UCAMWEBAUTH_RETURN_URL=u):
            req = (RequestFactory().get(reverse('raven_return')))
            self.assertEqual(get_return_url(req), u)
            self.client.get(reverse('raven_return'),
                            {'WLS-Response': create_wls_response()})
        self.assertIn('_auth_user_id', self.client.session)

    def test_params(self):
        testparams = {'this': ['that%21%25!+/']}
        raw = self.get_wls_response(
            raven_params=urlencode(testparams, doseq=True))
        r = RavenResponse(RequestFactory().get(reverse('raven_return'),
                                               {'WLS-Response': raw}))
        self.assertEqual(r.params, testparams)

    def test_get_next(self):
        testparams = {'next': ['http://foo.example/!++!%2F/']}
        raw = self.get_wls_response(
            raven_params=urlencode(testparams, doseq=True))
        next_p = get_next_from_wls_response(raw)
        self.assertEqual(next_p, testparams['next'][0])

    def test_empty_next(self):
        response = self.client.get('/accounts/login/')
        self.assertEqual(response.status_code, 303)
        self.assertTrue(response.url.startswith(settings.UCAMWEBAUTH_LOGIN_URL))

    def test_next_param(self):
        testnext = 'http://foo.example/!++!%2F/'
        response = self.client.get(reverse('raven_login'), {'next': testnext})
        self.assertEqual(
            parse_qs(parse_qs(urlparse(response['Location']).query)
                     ['params'][0])['next'][0],
            testnext)

    def test_exception_trace(self):
        # This is a bit fragile.  The aim is to check that RavenAuthBackend
        # isn't overwriting the traceback of exceptions raised by
        # RavenResponse.
        try:
            # Passing no args causes an exception.
            RavenAuthBackend().authenticate()
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            found = False
            while exc_traceback is not None:
                # Check that the traceback reaches RavenResponse.__init__
                if exc_traceback.tb_frame.f_code == RavenResponse.__init__.__code__:
                    found = True
                exc_traceback = exc_traceback.tb_next
            self.assertTrue(found)



import time
try:
    from urlparse import parse_qs
    from urllib import unquote
except ImportError:
    from urllib.parse import parse_qs, unquote
from OpenSSL.crypto import FILETYPE_PEM, load_certificate, verify
from django.conf import settings
from ucamwebauth.utils import decode_sig, setting, parse_time, get_return_url
from ucamwebauth.exceptions import (MalformedResponseError, InvalidResponseError, PublicKeyNotFoundError,
                                    UserNotAuthorised, OtherStatusCode)


class RavenResponse(object):
    """Transforms a WLS-Response (http://raven.cam.ac.uk/project/waa2wls-protocol.txt) from the
    University of Cambridge web login service (WLS) a.k.a. Raven (http://raven.cam.ac.uk/) into an object with
    accessible variables corresponding to the response parameters"""

    ver = status = msg = issue = ident = url = principal = ptags = auth = sso = life = params = kid = sig = None

    STATUS = {200: 'Successful authentication',
              410: 'The user cancelled the authentication request',
              510: 'No mutually acceptable authentication types available',
              520: 'Unsupported protocol version',
              530: 'General request parameter error',
              540: 'Interaction would be required',
              560: 'WAA not authorised',
              570: 'Authentication declined'}

    def __init__(self, response_req=None):
        """Creates a RavenResponse object from the response of the Web login service (WLS) of the University of
        Cambridge
        @param response_req The HTTP request that contains the WLS response.
        """

        if response_req is None:
            raise TypeError("no request supplied")
        try:
            response_str = response_req.GET['WLS-Response']
        except KeyError:
            raise TypeError("no WLS-Response")

        # The WLS sends an authentication response message as follows:  First a 'encoded response string' is formed by
        # concatenating the values of the response fields below, in the order shown, using '!' as a separator character.
        # If the characters '!'  or '%' appear in any
        # field value they MUST be replaced by their %-encoded representation
        # before concatenation.
        # Parameters with no relevant value MUST be encoded as the empty string.
        rawtokens = response_str.split('!')
        tokens = list(map(unquote, rawtokens))  # return a list for python3 compatibility

        # ver: The version of the WLS protocol in use. May be the same as the 'ver' parameter
        # supplied in the request
        try:
            self.ver = int(tokens[0])
        except ValueError:
            raise MalformedResponseError("Version number must be an integer, not %s" % tokens[0])
        if not 4 > self.ver > 0:
            raise MalformedResponseError("Unsupported version: %d" % self.ver)

        if self.ver == 3:
            versioni = 0
        else:
            versioni = 1

        # Check that the number of parameters in the response is correct
        if len(tokens) != (14-versioni):
            raise MalformedResponseError("Wrong number of parameters in response: expected %d, got %d" %
                                         ((14-versioni), len(tokens)))

        # status: A three digit status code indicating the status of the authentication request. The list of possible
        # statuses can be seen in the STATUS dict of the RavenResponse object.
        try:
            self.status = int(tokens[1])
        except ValueError:
            raise MalformedResponseError("Status code must be an integer, not %s" % tokens[1])

        if self.status not in self.STATUS:
            raise InvalidResponseError("Status returned not known")

        # msg (optional): A text message further describing the status of the authentication request,
        # suitable for display to end-user.
        self.msg = tokens[2]

        # issue: The date and time that the authentication response was created.
        try:
            self.issue = parse_time(tokens[3])
        except ValueError:
            raise MalformedResponseError("Issue time is not a valid time, got %s" % tokens[3])

        # Check that the response is recent by comparing 'issue' with the current time. The WLS MUST and the WAA SHOULD
        # have their clocks synchronised by NTP or a similar mechanism. Providing the WAA has access to an
        # NTP-synchronised clock then allowing for a transmission time of 30-60 seconds is probably appropriate.
        # Otherwise allowance must be made for the maximum expected clock skew.
        if self.issue > time.time():
            raise InvalidResponseError("The timestamp on the response is in the future")
        if self.issue < time.time() - setting('UCAMWEBAUTH_TIMEOUT', 30):
            raise InvalidResponseError("Response has timed out - issued %s, now %s" %
                                       (time.asctime(time.gmtime(self.issue)), time.asctime()))

        # ident: An identifier for this response. 'ident', combined with 'issue' provides a uid for this response.
        self.ident = tokens[4]

        if self.ident == "":
            raise MalformedResponseError("Empty ID")

        # url: The value of url supplied in the authentication request and used to form the authentication response.
        self.url = tokens[5]

        # Check that 'url' represents the resource currently being
        # accessed.  The request has already been checked against
        # ALLOWED_HOSTS by Django.
        if self.url != get_return_url(response_req):
            raise InvalidResponseError("The URL in the response does not match the URL expected")

        # principal: Only present if status == 200, indicates the authenticated identity of the user
        if self.status == 200:
            if tokens[6] != "":
                self.principal = tokens[6]
            else:
                raise InvalidResponseError("The username is not present in the WLS response")
        else:
            if tokens[6] != "":
                raise InvalidResponseError("The username should not be present if the status code is not 200")

        # ptags (optional): A potentially empty sequence of text tokens separated by ',' indicating attributes
        # or properties of the identified principal. Possible values of this tag are not standardised and are
        # a matter for local definition by individual WLS operators (see note below). Web application agent (WAA)
        # SHOULD ignore values that they do not recognise.
        if versioni == 0:
            self.ptags = tokens[7].split(',')

        # auth (not-empty only if authentication was successfully established by interaction with the user):
        # This indicates which authentication type was used. v3 only supports 'pwd'
        self.auth = tokens[8-versioni]

        # sso (not-empty only if 'auth' is empty): Authentication must have been established based on previous
        # successful authentication interaction(s) with the user. This indicates which authentication types were used
        # on these occasions. This value consists of a sequence of text tokens as described below, separated by ','.
        self.sso = tokens[9-versioni].split(',')

        # life (optional): If the user has established an authenticated 'session' with the WLS, this indicates the
        # remaining life (in seconds) of that session. If present, a WAA SHOULD use this to establish an upper limit
        # to the lifetime of any session that it establishes.
        # TODO https://docs.djangoproject.com/en/dev/topics/http/sessions/#django.contrib.sessions.backends.base.SessionBase.set_expiry
        if tokens[10-versioni] != "":
            try:
                self.life = int(tokens[10-versioni])
            except ValueError:
                raise MalformedResponseError("Life parameter must be an integer, not %s" % tokens[10-versioni])

        # params: a copy of the params parameter from the request
        try:
            self.params = parse_qs(tokens[11-versioni])
        except Exception:
            raise MalformedResponseError("The params field contains wrong characters: %s" % tokens[11-versioni])

        # REQUIRED to be a copy of the params parameter from the request
        # if self.params != setting('UCAMWEBAUTH_PARAMS', default=''):
        #     raise InvalidResponseError("The params are not equals to the request ones")

        # kid (not-empty only if 'sig' is present): A string which identifies the RSA key which was used to form the
        # signature supplied with the response. Typically these will be small integers.
        if tokens[12-versioni] != "":
            try:
                self.kid = int(tokens[12-versioni])
            except ValueError:
                raise MalformedResponseError("kid parameter must be an integer, not %s" % tokens[12-versioni])

        # sig (not-empty only if 'status' is 200): A public-key signature of the response data constructed from the
        # entire parameter value except 'kid' and 'sig' (and their separating ':' characters) using the private key
        # identified by 'kid', the SHA-1 hash algorithm and the 'RSASSA-PKCS1-v1_5' scheme as specified in PKCS #1 v2.1
        # [RFC 3447] and the resulting signature encoded using the base64 scheme [RFC 1521] except that the
        # characters '+', '/', and '=' are replaced by '-', '.' and '_' to reduce the URL-encoding overhead.
        if tokens[13-versioni] != "":
            if self.kid is None:
                raise InvalidResponseError("kid must be present if signature is present")
            self.sig = decode_sig(tokens[13-versioni])
        else:
            if self.status == 200:
                raise InvalidResponseError("Signature must be present if status is 200")

        # Check that 'kid', corresponds to a key/certificate present in the WAA. Is the only way to check the
        # signature. The WAA has to use the public key/certificate made available by the WLS.
        if (self.sig is not None) or (self.status == 200):
            try:
                cert = load_certificate(FILETYPE_PEM, settings.UCAMWEBAUTH_CERTS[self.kid])
            except Exception:
                raise PublicKeyNotFoundError("The server do not have the public key corresponding to the key the web "
                                             "login service signed the response with")

            # Check that the signature matches the data supplied. To check this, the WAA uses the public key identified
            # by 'kid'.
            data = '!'.join(rawtokens[0:(12-versioni)])
            # The data string that was signed in the WLS (everything from the  WLS-Response except 'kid' and 'sig'
            try:
                verify(cert, self.sig, data.encode(), 'sha1')
            except Exception:
                raise InvalidResponseError("The signature for this response is not valid.")

        if self.status == 200:

            # Check that 'auth' and/or 'sso' contain values acceptable to the WAA. Simply setting 'aauth' and 'iact'
            # values in an authentication request is not sufficient since an attacker could construct its own request.
            # Conversely, the WAA MUST ensure that the values of 'aauth' and/or 'iact' in its authentication requests
            # correctly reflect its requirement, to prevent the WLS sending it unacceptable responses.

            UCAMWEBAUTH_IACT = setting('UCAMWEBAUTH_IACT', '')

            # the authentication was successfully establish by interaction with the user
            if self.auth != "":
                # auth only supports 'pwd' in current version, therefore we compare it with 'pwd' only
                # If more are supported in the future, a setting will be added to specify which ones the WAA wants to
                # support and check that auth and sso match any element in this list.
                if self.auth != "pwd":
                    raise InvalidResponseError("The response used the wrong type of authentication (auth)")

                if UCAMWEBAUTH_IACT == 'no':
                    # We had required a non-interactive authentication, but didn't get one
                    raise InvalidResponseError("Non-interactive authentication required but not received")

            # authentication was established on a previous interaction(s) with the user
            else:
                if self.sso != [""]:
                    if self.sso != ["pwd"]:
                        raise InvalidResponseError("The response used the wrong type of authentication (sso)")

                    if UCAMWEBAUTH_IACT == 'yes':
                        # We had required an interactive authentication, but didn't get one
                        raise InvalidResponseError("Interactive authentication required but not received")
                else:
                    # Both auth and sso are empty, which is not allowed
                    raise MalformedResponseError("No authentication types supplied")

    def validate(self):
        """Returns True if this represents a successful authentication otherwise returns False."""
        return self.status == 200



import time
import calendar
from base64 import b64decode
try:
    from urlparse import parse_qs
    from urllib import unquote
except ImportError:
    from urllib.parse import parse_qs, unquote
from django.conf import settings
from django.http import HttpResponseRedirect
from django.core.urlresolvers import reverse
from ucamwebauth.exceptions import MalformedResponseError


def decode_sig(sig):
    """Decodes a signature from the variant base64 used by raven.
    @param sig  A string giving the signature in Raven's variant base-64
    @return  A binary string containing the signature"""
    sig = sig.replace("-", "+").replace(".", "/").replace("_", "=")
    try:
        return b64decode(sig)
    except TypeError:
        raise MalformedResponseError("Signature is not a valid base-64 encoded string")


def setting(name, default=None):
    """Returns a setting from the Django settings file"""
    return getattr(settings, name, default)


def parse_time(time_string):
    """Converts a time of the form '20110729T123456Z' to a number of seconds
    since the epoch.
    @exception ValueError if the time is not a valid Raven time"""
    return calendar.timegm(time.strptime(time_string, "%Y%m%dT%H%M%SZ"))


def get_next_from_wls_response(response_str):
    """ Returns the value of the variable 'next' inside the parameter 'params' of the response
    :param response_str: The WLS response
    :return: the value of the 'next' variable
    """
    tokens = list(map(unquote, response_str.split('!')))
    params = parse_qs(tokens[11]) if tokens[0] == '3' else parse_qs(tokens[10])
    if 'next' in params:
        return params['next'][0]
    else:
        return None


def get_return_url(request):
    """Generate the return URL for a particular request (either a request
    that needs to be authenticated, or one that contains a purported
    authentication response)."""
    return setting('UCAMWEBAUTH_RETURN_URL',
                   default=request.build_absolute_uri(reverse('raven_return')))


class HttpResponseSeeOther(HttpResponseRedirect):
    """An HttpResponse with a 303 status code, since django doesn't provide one
    by default.  A 303 is required by the the WAA2WLS specification."""
    status_code = 303
